@@ -14,6 +14,30 @@
 
 ;;; Utilities
 
+(defun magit-annex-tests-kill-buffers (directory)
+  (let ((default-directory directory))
+    (dolist (buf (-remove #'buffer-base-buffer (magit-mode-get-buffers)))
+      (kill-buffer buf))))
+
+(defmacro magit-annex-tests-wait (&rest body)
+  (declare (indent 0) (debug t))
+  (let ((wait (gensym "magit-annex-tests-wait")))
+    `(prog1 ,(macroexp-progn body)
+       (let ((,wait t))
+         (when magit-this-process
+           (set-process-sentinel
+            magit-this-process
+            (lambda (process event)
+              (magit-process-sentinel process event)
+              (when (memq (process-status process) '(exit signal))
+                (setq ,wait nil))))
+           (let ((start (float-time)))
+             (while (and magit-this-process ,wait)
+               (when (< 30 (time-to-seconds
+                            (subtract-time (float-time) start)))
+                 (error "Process shouldn't take this long"))
+               (sleep-for 0.005))))))))
+
 ;; Modified from Magit's magit-with-test-directory.
 (defmacro magit-annex-with-test-directory (&rest body)
   (declare (indent 0) (debug t))
@@ -23,11 +47,12 @@
        (push "GIT_AUTHOR_NAME=A U Thor" process-environment)
        (push "GIT_AUTHOR_EMAIL=a.u.thor@example.com" process-environment)
        (condition-case err
-           (cl-letf (((symbol-function #'message) (lambda (&rest _))))
+           (cl-letf (((symbol-function #'message) #'format))
              (let ((default-directory ,dir))
                ,@body))
-         (error (message "Keeping test directory:\n  %s" ,dir)
+         (error (message "Keeping test directory and buffers:\n  %s" ,dir)
                 (signal (car err) (cdr err))))
+       (magit-annex-tests-kill-buffers ,dir)
        (delete-directory ,dir t))))
 
 (defmacro magit-annex-with-test-repo (&rest body)
@@ -47,7 +72,7 @@
      (push "GIT_AUTHOR_NAME=A U Thor" process-environment)
      (push "GIT_AUTHOR_EMAIL=a.u.thor@example.com" process-environment)
      (condition-case err
-         (cl-letf (((symbol-function #'message) (lambda (&rest _))))
+         (cl-letf (((symbol-function #'message) #'format))
            (let ((default-directory repo1))
              (magit-call-git "init" ".")
              (magit-call-git "annex" "init" "repo1")
@@ -60,41 +85,17 @@
              (magit-call-git "annex" "init" "repo2"))
            (let ((default-directory repo1))
              ,@body))
-       (error (message "Keeping test directories:\n  %s\n%s" repo1 repo2)
-              (signal (car err) (cdr err))))
+       (error
+        (message "Keeping test directories and buffers:\n  %s\n  %s"
+                 repo1 repo2)
+        (signal (car err) (cdr err))))
+     (magit-annex-tests-kill-buffers repo1)
+     (magit-annex-tests-kill-buffers repo2)
      (call-process "chmod" nil nil nil "-R" "777" repo1)
      (call-process "chmod" nil nil nil "-R" "777" repo2)
      (delete-directory repo1 t)
      (delete-directory repo2 t)))
 
-(defmacro magit-annex-tests-with-temp-bare-repo (&rest body)
-  (declare (indent 0) (debug t))
-  `(magit-annex-with-test-directory
-     (magit-call-git "init" "-bare" ".")
-     ,@body))
-
-(defmacro magit-annex-tests-with-temp-clone (url &rest body)
-  (declare (indent 1) (debug t))
-  (let ((repo (gensym)))
-    `(let ((,repo ,(or url 'default-directory)))
-       (magit-annex-with-test-directory
-         (magit-call-git "clone" ,repo ".")
-         (magit-call-git "annex" "init" "test-repo")
-         ;; Make a normal commit and push.
-         (magit-annex-tests-modify-file "file")
-         (magit-stage-file "file")
-         (magit-call-git "commit" "-m" "normal commit")
-         (magit-call-git "push")
-         (magit-call-git "push" "-u" "origin" "git-annex")
-         (unwind-protect
-             (progn ,@body)
-           (call-process "chmod" nil nil nil "-R" "777" "."))))))
-
-(defun magit-annex-tests-wait ()
-  (while (and magit-this-process
-              (eq (process-status magit-this-process) 'run))
-    (sleep-for 0.005)))
-
 (defun magit-annex-tests-modify-file (filename)
   (with-temp-file (expand-file-name filename)
     (insert (symbol-name (gensym "content")))))
@@ -102,9 +103,9 @@
 (defun magit-annex-tests-should-have-section (type info)
   (magit-status default-directory)
   (message (buffer-string))
-  (should (--first (equal (magit-section-value it) info)
-                   (magit-section-children
-                    (magit-get-section `((,type) (status)))))))
+  (should (--first (equal (oref it value) info)
+                   (oref (magit-get-section `((,type) (status)))
+                         children))))
 
 
 ;;; Annexing
@@ -138,14 +139,14 @@
       (magit-annex-tests-modify-file "annex-file")
       (magit-annex-add "annex-file")
       (magit-call-git "commit" "-m" "annex commit")
-      (magit-annex-sync)
-      (magit-annex-tests-wait)
+      (magit-annex-tests-wait
+        (magit-annex-sync))
       (should (magit-git-lines "diff" "repo1/master"))
       (should-not (magit-git-lines "diff" "synced/master"))
       (should (magit-annex-present-files)))
     (let ((default-directory repo1))
-      (magit-annex-merge)
-      (magit-annex-tests-wait)
+      (magit-annex-tests-wait
+        (magit-annex-merge))
       (should-not (magit-annex-present-files)))))
 
 (ert-deftest magit-annex-sync-content ()
@@ -154,14 +155,14 @@
       (magit-annex-tests-modify-file "annex-file")
       (magit-annex-add "annex-file")
       (magit-call-git "commit" "-m" "annex commit")
-      (magit-annex-sync '("--content"))
-      (magit-annex-tests-wait)
+      (magit-annex-tests-wait
+        (magit-annex-sync '("--content")))
       (should (magit-git-lines "diff" "repo1/master"))
       (should-not (magit-git-lines "diff" "synced/master"))
       (should (magit-annex-present-files)))
     (let ((default-directory repo1))
-      (magit-annex-merge)
-      (magit-annex-tests-wait)
+      (magit-annex-tests-wait
+        (magit-annex-merge))
       (should (magit-annex-present-files)))))
 
 
@@ -173,13 +174,13 @@
       (magit-annex-tests-modify-file "annex-file")
       (magit-annex-add "annex-file")
       (magit-call-git "commit" "-m" "annex commit")
-      (magit-annex-sync)
-      (magit-annex-tests-wait))
+      (magit-annex-tests-wait
+        (magit-annex-sync)))
     (let ((default-directory repo1))
-      (magit-annex-merge)
-      (magit-annex-tests-wait)
-      (magit-annex-get-all-auto)
-      (magit-annex-tests-wait)
+      (magit-annex-tests-wait
+        (magit-annex-merge))
+      (magit-annex-tests-wait
+        (magit-annex-get-all-auto))
       ;; Shouldn't be present because of --auto flag.
       (should-not (magit-annex-present-files)))))
 
@@ -189,27 +190,46 @@
       (magit-annex-tests-modify-file "annex-file")
       (magit-annex-add "annex-file")
       (magit-call-git "commit" "-m" "annex commit")
-      (magit-annex-sync)
-      (magit-annex-tests-wait))
+      (magit-annex-tests-wait
+        (magit-annex-sync)))
     (let ((default-directory repo1))
-      (magit-annex-merge)
-      (magit-annex-tests-wait)
+      (magit-annex-tests-wait
+        (magit-annex-merge))
       (should-not (magit-annex-present-files))
-      (magit-annex-get-files '("annex-file"))
-      (magit-annex-tests-wait)
+      (magit-annex-tests-wait
+        (magit-annex-get-files '("annex-file")))
       (should (equal (magit-annex-present-files)
                      '("annex-file"))))))
 
+(ert-deftest magit-annex-get-files-subdir ()
+  (magit-annex-with-test-repo-pair
+    (let ((default-directory repo2))
+      (make-directory "subdir")
+      (magit-annex-tests-modify-file "subdir/annex-file")
+      (magit-annex-add "subdir/annex-file")
+      (magit-call-git "commit" "-m" "subdir annex commit")
+      (magit-annex-tests-wait
+        (magit-annex-sync)))
+    (let ((default-directory repo1))
+      (magit-annex-tests-wait
+        (magit-annex-merge))
+      (should-not (magit-annex-present-files))
+      (let ((default-directory (concat repo1 "subdir")))
+        (magit-annex-tests-wait
+          (magit-annex-get-files '("annex-file"))))
+      (should (equal (magit-annex-present-files)
+                     '("subdir/annex-file"))))))
+
 (ert-deftest magit-annex-drop-files ()
   (magit-annex-with-test-repo-pair
     (let ((default-directory repo2))
       (magit-annex-tests-modify-file "annex-file")
       (magit-annex-add "annex-file")
       (magit-call-git "commit" "-m" "annex commit")
-      (magit-annex-sync)
-      (magit-annex-tests-wait)
-      (magit-annex-drop-files '("annex-file") '("--force"))
-      (magit-annex-tests-wait)
+      (magit-annex-tests-wait
+        (magit-annex-sync))
+      (magit-annex-tests-wait
+        (magit-annex-drop-files '("annex-file") '("--force")))
       (should-not (magit-annex-present-files)))))
 
 (ert-deftest magit-annex-move-files ()
@@ -218,13 +238,14 @@
       (magit-annex-tests-modify-file "annex-file")
       (magit-annex-add "annex-file")
       (magit-call-git "commit" "-m" "annex commit")
-      (magit-annex-sync)
-      (magit-annex-tests-wait)
-      (magit-annex-move-files '("annex-file") '("--to=repo1"))
-      (magit-annex-tests-wait)
+      (magit-annex-tests-wait
+        (magit-annex-sync))
+      (magit-annex-tests-wait
+        (magit-annex-move-files '("annex-file") '("--to=repo1")))
       (should-not (magit-annex-present-files)))
     (let ((default-directory repo1))
-      (magit-annex-merge)
+      (magit-annex-tests-wait
+        (magit-annex-merge))
       (should (equal (magit-annex-present-files)
                      '("annex-file"))))))
 
@@ -234,14 +255,15 @@
       (magit-annex-tests-modify-file "annex-file")
       (magit-annex-add "annex-file")
       (magit-call-git "commit" "-m" "annex commit")
-      (magit-annex-sync)
-      (magit-annex-tests-wait)
-      (magit-annex-copy-files '("annex-file") '("--to=repo1"))
-      (magit-annex-tests-wait)
+      (magit-annex-tests-wait
+        (magit-annex-sync))
+      (magit-annex-tests-wait
+        (magit-annex-copy-files '("annex-file") '("--to=repo1")))
       (should (equal (magit-annex-present-files)
                      '("annex-file"))))
     (let ((default-directory repo1))
-      (magit-annex-merge)
+      (magit-annex-tests-wait
+        (magit-annex-merge))
       (should (equal (magit-annex-present-files)
                      '("annex-file"))))))
 
@@ -252,9 +274,23 @@
     (magit-call-git "commit" "-m" "annex commit")
     (should-not (magit-annex-unlocked-files))
     (magit-annex-unlock-files '("annex-file"))
-    (magit-annex-tests-wait)
     (should (equal (magit-annex-unlocked-files)
                    '("annex-file")))
     (magit-annex-lock-files '("annex-file") '("--force"))
-    (magit-annex-tests-wait)
     (should-not (magit-annex-unlocked-files))))
+
+(ert-deftest magit-annex-undo-files ()
+  (magit-annex-with-test-repo
+    (magit-annex-tests-modify-file "annex-file")
+    (magit-annex-add "annex-file")
+    (magit-call-git "commit" "-m" "annex commit")
+    (let* ((key-fn (lambda (x)
+                     (magit-git-str "annex" "lookupkey" x)))
+           (orig-key (funcall key-fn "annex-file") ))
+      (magit-annex-unlock-files '("annex-file"))
+      (magit-annex-tests-modify-file "annex-file")
+      (magit-annex-add "annex-file")
+      (magit-call-git "commit" "-m" "update file")
+      (should-not (equal orig-key (funcall key-fn "annex-file")))
+      (magit-annex-undo-files '("annex-file"))
+      (should (equal orig-key (funcall key-fn "annex-file"))))))
