@@ -6,8 +6,8 @@
 ;;         Rémi Vanicat <vanicat@debian.org>
 ;; URL: https://github.com/magit/magit-annex
 ;; Keywords: vc tools
-;; Version: 1.6.0
-;; Package-Requires: ((cl-lib "0.3") (magit "2.12.0"))
+;; Version: 1.7.1
+;; Package-Requires: ((cl-lib "0.3") (magit "2.90.0"))
 
 ;; This program is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -98,15 +98,15 @@
   "Whether to confirm before adding all changes to the annex."
   :type 'boolean)
 
+(define-obsolete-variable-alias 'magit-annex-standard-options
+  'magit-annex-global-arguments "1.4.0")
+
 (defcustom magit-annex-global-arguments nil
   "Arguments that are added to every git-annex call.
 These are placed after \"annex\" in the call, whereas values from
 `magit-git-global-arguments' are placed after \"git\"."
   :type '(repeat string))
 
-(define-obsolete-variable-alias 'magit-annex-standard-options
-  'magit-annex-global-arguments "1.4.0")
-
 (defcustom magit-annex-limit-file-choices t
   "Limit choices for file commands based on state of repo.
 For example, if locking a file, limit choices to unlocked files."
@@ -231,9 +231,11 @@ program used to open the unused file."
 ;;;###autoload
 (eval-after-load 'magit
   '(progn
-     (define-key magit-mode-map "@" 'magit-annex-popup-or-init)
-     (magit-define-popup-action 'magit-dispatch-popup
-       ?@ "Annex" 'magit-annex-popup-or-init ?!)))
+     (require 'magit-popup)
+     (when (boundp 'magit-dispatch-popup)
+       (define-key magit-mode-map "@" 'magit-annex-popup-or-init)
+       (magit-define-popup-action 'magit-dispatch-popup
+         ?@ "Annex" 'magit-annex-popup-or-init ?!))))
 
 
 
@@ -412,18 +414,17 @@ With a prefix argument, prompt for FILE.
               (capitalize command) command)
      (interactive
       (list
-       (let ((default
-               (--when-let
-                   (or (mapcar #'cdr (magit-region-values 'annex-list-file))
-                       (-some-> (cdr (magit-section-when annex-list-file))
-                                (list))
-                       (and (derived-mode-p 'dired-mode)
-                            (dired-get-marked-files t)))
-                 (mapconcat #'identity it ","))))
+       (let* ((files (or (mapcar #'cdr (magit-region-values 'annex-list-file))
+                         (when-let ((file (cdr (magit-section-value-if
+                                                'annex-list-file))))
+                           (list file))
+                         (and (derived-mode-p 'dired-mode)
+                              (dired-get-marked-files t))))
+              (default (mapconcat #'identity files ",")))
          (magit-annex-read-files
           (concat ,(capitalize command)
                   " file,s"
-                  (and default (format " (%s)" default))
+                  (and files (format " (%s)" default))
                   ": ")
           ,limit
           default))
@@ -530,8 +531,7 @@ value for the '-S' flag.  The '--stat' flag is also enabled if
         (magit-invoke-popup 'magit-log-popup nil nil)))))
 
 (defun magit-annex--file-name-from-key (key)
-  (magit-git-string "annex" "examinekey" key
-                    "--format=.git/annex/objects/${hashdirmixed}${key}/${key}"))
+  (magit-git-string "annex" "contentlocation" key))
 
 (declare-function dired-read-shell-command "dired-aux" (prompt arg files))
 
